/**
 * A specialization of `Program` with the capability of mapping `Model` to `View`
 * and rendering it into a DOM node.
 *
 * `Html` is a base abstraction in order to work with any library that renders html.
 *
 * @since 0.5.0
 */
import { map as RxMap, takeUntil } from 'rxjs/operators';
import * as platform from './Platform';
import { none } from './Sub';
/**
 * Maps a view which carries a message of type `A` into a view which carries a message of type `B`.
 * @category Functor
 * @since 0.5.0
 */
export function map(f) {
    return ha => dispatch => ha(a => dispatch(f(a)));
}
/**
 * Returns a `Program` specialized for `Html`.
 *
 * It needs a `view()` function that maps `Model` to `Html`.
 *
 * Underneath it uses `Platform.program()`.
 * @category constructors
 * @since 0.5.0
 */
export function program(init, update, view, subscriptions = () => none) {
    const { dispatch, cmd$, sub$, model$ } = platform.program(init, update, subscriptions);
    const html$ = model$.pipe(RxMap(view));
    return { dispatch, cmd$, sub$, model$, html$ };
}
/**
 * Same as `program()` but with `Flags` that can be passed when the `Program` is created in order to manage initial values.
 * @category constructors
 * @since 0.5.0
 */
export function programWithFlags(init, update, view, subscriptions) {
    return flags => program(init(flags), update, view, subscriptions);
}
/**
 * Stops the `program` when `signal` Observable emits a value.
 * @category combinators
 * @since 0.5.4
 */
export function withStop(signal) {
    return program => {
        const platformProgram = platform.withStop(signal)(program);
        return Object.assign(Object.assign({}, platformProgram), { html$: program.html$.pipe(takeUntil(signal)) });
    };
}
/**
 * Runs the `Program`.
 *
 * Underneath it uses `Platform.run()`.
 *
 * It subscribes to the views stream (`html$`) and runs `Renderer` for each new value.
 * @category utils
 * @since 0.5.0
 */
export function run(program, renderer) {
    const { dispatch, html$ } = program;
    html$.subscribe(html => renderer(html(dispatch)));
    return platform.run(program);
}
