/**
 * Makes http calls to remote resources as `Cmd`s.
 *
 * See [Http](https://package.elm-lang.org/packages/elm/http/latest/Http) Elm package.
 *
 * @since 0.5.0
 */
import * as Arr from 'fp-ts/es6/Array';
import * as E from 'fp-ts/es6/Either';
import * as O from 'fp-ts/es6/Option';
import * as Rec from 'fp-ts/es6/Record';
import { getLastSemigroup } from 'fp-ts/es6/Semigroup';
import * as T from 'fp-ts/es6/Task';
import { flow } from 'fp-ts/es6/function';
import { pipe } from 'fp-ts/es6/pipeable';
import { of } from 'rxjs';
import { AjaxError, AjaxTimeoutError, ajax } from 'rxjs/ajax';
import { catchError, map } from 'rxjs/operators';
/**
 * @category destructors
 * @since 0.5.0
 */
export function toTask(req) {
    return () => xhrOnlyBody(req).toPromise();
}
/**
 * Executes as `Cmd` the provided call to remote resource, mapping the result to a `Msg`.
 *
 * Derived from [`sendBy`](#sendBy).
 *
 * @since 0.5.0
 */
export function send(f) {
    return sendBy(flow(E.map(r => r.body), f));
}
/**
 * Executes as `Cmd` the provided call to remote resource, mapping the full Response to a `Msg`.
 *
 * @since 0.6.0
 */
export function sendBy(f) {
    return flow(xhr, toMsg(f));
}
/**
 * @category constructors
 * @since 0.5.0
 */
export function get(url, decoder) {
    return {
        method: 'GET',
        headers: {},
        url,
        body: undefined,
        expect: decoder,
        timeout: O.none,
        withCredentials: false
    };
}
/**
 * @category constructors
 * @since 0.5.0
 */
export function post(url, body, decoder) {
    return {
        method: 'POST',
        headers: {},
        url,
        body,
        expect: decoder,
        timeout: O.none,
        withCredentials: false
    };
}
const fromStrArr = Rec.fromFoldableMap(getLastSemigroup(), Arr.array);
const xhrOnlyBody = flow(xhr, extractBody());
function toMsg(project) {
    return map(flow(project, O.some, T.of));
}
function extractBody() {
    return map(E.map(response => response.body));
}
function xhr(req) {
    return pipe(toXHRRequest(req), ajax, map(flow(toResponse, decodeWith(req.expect))), catchError((e) => of(E.left(toHttpError(req, e)))));
}
function toXHRRequest(req) {
    return Object.assign(Object.assign({}, req), { timeout: O.getOrElse(() => 0)(req.timeout), async: true, responseType: 'text' });
}
function toResponse(resp) {
    return {
        url: resp.request.url,
        status: { code: resp.status, message: resp.xhr.statusText },
        headers: toResponseHeaders(resp.xhr),
        body: typeof resp.response === 'string' && resp.response.length > 0 ? resp.response : '{}'
    };
}
function decodeWith(decoder) {
    return response => pipe(
    // By spec parsing json can only throw `SyntaxError`
    E.parseJSON(response.body, e => e.message), E.chain(decoder), E.bimap(value => ({ _tag: 'BadPayload', value, response }), body => (Object.assign(Object.assign({}, response), { body }))));
}
function toHttpError(req, e) {
    if (e instanceof AjaxTimeoutError) {
        return { _tag: 'Timeout' };
    }
    if (e instanceof AjaxError && e.status === 404) {
        return { _tag: 'BadUrl', value: req.url };
    }
    if (e instanceof AjaxError && e.status !== 0) {
        return {
            _tag: 'BadStatus',
            response: {
                url: e.request.url,
                status: { code: e.status, message: e.xhr.statusText },
                headers: toResponseHeaders(e.xhr),
                body: e.response
            }
        };
    }
    return { _tag: 'NetworkError', value: e instanceof Error ? e.message : '' };
}
function toResponseHeaders(xhr) {
    return pipe(xhr.getAllResponseHeaders(), O.fromPredicate((x) => x.length > 0), O.map(hs => hs.trim().split(/[\r\n]+/)), O.map(Arr.map(line => line.split(': '))), O.map(kvs => fromStrArr(kvs, ([k, v]) => [k, v])), O.getOrElse(() => ({})));
}
