/**
 * A specialization of `Program` that handles application navigation via location's hash.
 *
 * It uses [`history`](https://github.com/ReactTraining/history) package.
 *
 * @since 0.5.0
 */
import * as O from 'fp-ts/es6/Option';
import * as H from 'history';
import { Subject, of } from 'rxjs';
import { map } from 'rxjs/operators';
import * as html from './Html';
import { batch, none } from './Sub';
const history = H.createHashHistory();
/**
 * Location changes are expressed as a stream
 */
const location$ = new Subject();
history.listen(location => {
    location$.next(location);
});
/**
 * Generates a `Cmd` that adds a new location to the history's list.
 * @category utils
 * @since 0.5.0
 */
export function push(url) {
    return of(() => {
        history.push(url);
        return Promise.resolve(O.none);
    });
}
/**
 * Returns a `Program` specialized for `Navigation`.
 *
 * The `Program` is a `Html.Program` but it needs a `locationToMsg()` function which converts location changes to messages.
 *
 * Underneath it consumes `location$` stream (applying `locationToMsg()` on its values).
 * @category constructors
 * @since 0.5.0
 */
export function program(locationToMessage, init, update, view, subscriptions = () => none) {
    const onChangeLocation$ = location$.pipe(map(location => locationToMessage(location)));
    const subs = (model) => batch([subscriptions(model), onChangeLocation$]);
    return html.program(init(history.location), update, view, subs);
}
/**
 * Same as `program()` but with `Flags` that can be passed when the `Program` is created in order to manage initial values.
 * @category constructors
 * @since 0.5.0
 */
export function programWithFlags(locationToMessage, init, update, view, subscriptions = () => none) {
    return flags => program(locationToMessage, init(flags), update, view, subscriptions);
}
