"use strict";
/**
 * This module makes available a debugging utility for `elm-ts` applications running `Html` programs.
 *
 * `elm-ts` ships with a [Redux DevTool Extension](https://github.com/zalmoxisus/redux-devtools-extension) integration, falling back to a simple debugger via standard browser's [`console`](https://developer.mozilla.org/en-US/docs/Web/API/Console) in case the extension is not available.
 *
 * **Note:** debugging is to be considered unsafe by design so it should be used only in **development**.
 *
 * This is an example of usage:
 * ```ts
 * import {react, cmd} from 'elm-ts'
 * import {programWithDebugger} from 'elm-ts/lib/Debug/Html'
 * import { render } from 'react-dom'
 *
 * type Model = number
 * type Msg = 'INCREMENT' | 'DECREMENT'
 *
 * declare const init: [Model, cmd.none]
 * declare function update(msg: Msg, model: Model): [Model, cmd.Cmd<Msg>]
 * declare function view(model: Model): react.Html<Msg>
 *
 * const program = process.NODE_ENV === 'production' ? react.program : programWithDebugger
 *
 * const main = program(init, update, view)
 *
 * react.run(main, dom => render(dom, document.getElementById('app')))
 * ```
 *
 * @since 0.5.3
 */
Object.defineProperty(exports, "__esModule", { value: true });
var rxjs_1 = require("rxjs");
var Html_1 = require("../Html");
var commons_1 = require("./commons");
/**
 * Adds a debugging capability to a generic `Html` `Program`.
 *
 * It tracks every `Message` dispatched and resulting `Model` update.
 *
 * It also lets directly updating the application's state with a special `Message` of type:
 *
 * ```ts
 * {
 *   type: '__DebugUpdateModel__'
 *   payload: Model
 * }
 * ```
 *
 * or applying a message with:
 * ```ts
 * {
 *   type: '__DebugApplyMsg__';
 *   payload: Msg
 * }
 * ```
 * @category constructors
 * @since 0.5.3
 */
function programWithDebugger(init, update, view, subscriptions) {
    return createProgram(rxjs_1.EMPTY, init, update, view, subscriptions);
}
exports.programWithDebugger = programWithDebugger;
/**
 * A function that requires an `Observable` and returns a `programWithDebugger()` function: the underlying debugger will stop when the `Observable` emits a value.
 * @category constructors
 * @since 0.5.4
 */
function programWithDebuggerWithStop(stopDebuggerOn) {
    return function (init, update, view, subscriptions) { return createProgram(stopDebuggerOn, init, update, view, subscriptions); };
}
exports.programWithDebuggerWithStop = programWithDebuggerWithStop;
/**
 * Same as `programWithDebugger()` but with `Flags` that can be passed when the `Program` is created in order to manage initial values.
 * @category constructors
 * @since 0.5.3
 */
function programWithDebuggerWithFlags(init, update, view, subscriptions) {
    return function (flags) { return programWithDebugger(init(flags), update, view, subscriptions); };
}
exports.programWithDebuggerWithFlags = programWithDebuggerWithFlags;
/**
 * Same as `programWithDebuggerWithStop()` but with `Flags` that can be passed when the `Program` is created in order to manage initial values.
 * @category constructors
 * @since 0.5.4
 */
function programWithDebuggerWithFlagsWithStop(stopDebuggerOn) {
    return function (init, update, view, subscriptions) { return function (flags) {
        return createProgram(stopDebuggerOn, init(flags), update, view, subscriptions);
    }; };
}
exports.programWithDebuggerWithFlagsWithStop = programWithDebuggerWithFlagsWithStop;
// --- Internal
function createProgram(stopDebuggerOn, init, update, view, subscriptions) {
    var Debugger = commons_1.runDebugger(window, stopDebuggerOn);
    var initModel = init[0];
    var debug$ = new rxjs_1.BehaviorSubject([commons_1.debugInit(), initModel]);
    var p = Html_1.program(init, commons_1.updateWithDebug(debug$, update), view, subscriptions);
    // --- Run the debugger
    // --- we need to make a type assertion for `dispatch` because we cannot change the intrinsic `msg` type of `program`;
    // --- otherwise `programWithDebugger` won't be usable as a transparent extension/substitution of `Html`'s programs
    Debugger({
        debug$: debug$,
        init: initModel,
        dispatch: p.dispatch
    })();
    return p;
}
