"use strict";
/**
 * This module makes available a debugging utility for `elm-ts` applications running `Navigation` programs.
 *
 * `elm-ts` ships with a [Redux DevTool Extension](https://github.com/zalmoxisus/redux-devtools-extension) integration, falling back to a simple debugger via standard browser's [`console`](https://developer.mozilla.org/en-US/docs/Web/API/Console) in case the extension is not available.
 *
 * **Note:** debugging is to be considered unsafe by design so it should be used only in **development**.
 *
 * This is an example of usage:
 * ```ts
 * import {react, cmd} from 'elm-ts'
 * import {programWithDebugger} from 'elm-ts/lib/Debug/Navigation'
 * import {Location, program} from 'elm-ts/lib/Navigation'
 * import {render} from 'react-dom'
 *
 * type Model = number
 * type Msg = 'INCREMENT' | 'DECREMENT'
 *
 * declare function locationToMsg(location: Location): Msg
 * declare function init(location: Location): [Model, cmd.none]
 * declare function update(msg: Msg, model: Model): [Model, cmd.Cmd<Msg>]
 * declare function view(model: Model): react.Html<Msg>
 *
 * const program = process.NODE_ENV === 'production' ? program : programWithDebugger
 *
 * const main = program(locationToMsg, init, update, view)
 *
 * react.run(main, dom => render(dom, document.getElementById('app')))
 * ```
 *
 * @since 0.5.3
 */
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
var H = __importStar(require("history"));
var rxjs_1 = require("rxjs");
var Navigation_1 = require("../Navigation");
var commons_1 = require("./commons");
/**
 * Adds a debugging capability to a generic `Navigation` `Program`.
 *
 * It tracks every `Message` dispatched and resulting `Model` update.
 *
 * It also lets directly updating the application's state with a special `Message` of type:
 *
 * ```ts
 * {
 *   type: '__DebugUpdateModel__'
 *   payload: Model
 * }
 * ```
 *
 * or applying a message with:
 * ```ts
 * {
 *   type: '__DebugApplyMsg__';
 *   payload: Msg
 * }
 * ```
 * @category constructors
 * @since 0.5.3
 */
function programWithDebugger(locationToMessage, init, update, view, subscriptions) {
    return createProgram(rxjs_1.EMPTY, locationToMessage, init, update, view, subscriptions);
}
exports.programWithDebugger = programWithDebugger;
/**
 * A function that requires an `Observable` and returns a `programWithDebugger()` function: the underlying debugger will stop when the `Observable` emits a value.
 * @category constructors
 * @since 0.5.4
 */
function programWithDebuggerWithStop(stopDebuggerOn) {
    return function (locationToMessage, init, update, view, subscriptions) {
        return createProgram(stopDebuggerOn, locationToMessage, init, update, view, subscriptions);
    };
}
exports.programWithDebuggerWithStop = programWithDebuggerWithStop;
/**
 * Same as `programWithDebugger()` but with `Flags` that can be passed when the `Program` is created in order to manage initial values.
 * @category constructors
 * @since 0.5.3
 */
function programWithDebuggerWithFlags(locationToMessage, init, update, view, subscriptions) {
    return function (flags) { return programWithDebugger(locationToMessage, init(flags), update, view, subscriptions); };
}
exports.programWithDebuggerWithFlags = programWithDebuggerWithFlags;
/**
 * Same as `programWithDebuggerWithStop()` but with `Flags` that can be passed when the `Program` is created in order to manage initial values.
 * @category constructors
 * @since 0.5.4
 */
function programWithDebuggerWithFlagsWithStop(stopDebuggerOn) {
    return function (locationToMessage, init, update, view, subscriptions) { return function (flags) {
        return createProgram(stopDebuggerOn, locationToMessage, init(flags), update, view, subscriptions);
    }; };
}
exports.programWithDebuggerWithFlagsWithStop = programWithDebuggerWithFlagsWithStop;
// --- Internal
function createProgram(stopDebuggerOn, locationToMessage, init, update, view, subscriptions) {
    var history = H.createHashHistory(); // this is needed only to generate init model for debug$ :S
    var Debugger = commons_1.runDebugger(window, stopDebuggerOn);
    var initModel = init(history.location)[0];
    var debug$ = new rxjs_1.BehaviorSubject([commons_1.debugInit(), initModel]);
    var p = Navigation_1.program(locationToMessage, init, commons_1.updateWithDebug(debug$, update), view, subscriptions);
    // --- Run the debugger
    // --- we need to make a type assertion for `dispatch` because we cannot change the intrinsic `msg` type of `program`;
    // --- otherwise `programWithDebugger` won't be usable as a transparent extension/substitution of `Html`'s programs
    Debugger({
        debug$: debug$,
        init: initModel,
        dispatch: p.dispatch
    })();
    return p;
}
