"use strict";
/**
 * The `Platform` module is the backbone of `elm-ts`.
 * It defines the base `program()` and `run()` functions which will be extended by more specialized modules.
 * _The Elm Architecture_ is implemented via **RxJS** `Observables`.
 *
 * @since 0.5.0
 */
var __assign = (this && this.__assign) || function () {
    __assign = Object.assign || function(t) {
        for (var s, i = 1, n = arguments.length; i < n; i++) {
            s = arguments[i];
            for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p))
                t[p] = s[p];
        }
        return t;
    };
    return __assign.apply(this, arguments);
};
var __rest = (this && this.__rest) || function (s, e) {
    var t = {};
    for (var p in s) if (Object.prototype.hasOwnProperty.call(s, p) && e.indexOf(p) < 0)
        t[p] = s[p];
    if (s != null && typeof Object.getOwnPropertySymbols === "function")
        for (var i = 0, p = Object.getOwnPropertySymbols(s); i < p.length; i++) {
            if (e.indexOf(p[i]) < 0 && Object.prototype.propertyIsEnumerable.call(s, p[i]))
                t[p[i]] = s[p[i]];
        }
    return t;
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
var O = __importStar(require("fp-ts/lib/Option"));
var rxjs_1 = require("rxjs");
var operators_1 = require("rxjs/operators");
var Sub_1 = require("./Sub");
/**
 * `program()` is the real core of `elm-ts`.
 *
 * When a new `Program` is defined, a `BehaviorSubject` is created (because an initial value is needed) that will track every change to the `Model` and every `Cmd` executed.
 *
 * Every time `dispatch()` is called a new value, computed by the `update()` function, is added to the the stream.
 * @category constructors
 * @since 0.5.0
 */
function program(init, update, subscriptions) {
    if (subscriptions === void 0) { subscriptions = function () { return Sub_1.none; }; }
    var state$ = new rxjs_1.BehaviorSubject(init);
    var dispatch = function (msg) { return state$.next(update(msg, state$.getValue()[0])); };
    var cmd$ = state$.pipe(operators_1.startWith(init), operators_1.map(function (state) { return state[1]; }), operators_1.distinctUntilChanged(), operators_1.mergeAll());
    var model$ = state$.pipe(operators_1.map(function (state) { return state[0]; }), operators_1.distinctUntilChanged());
    var sub$ = model$.pipe(operators_1.switchMap(function (model) { return subscriptions(model); }));
    return { dispatch: dispatch, cmd$: cmd$, sub$: sub$, model$: model$ };
}
exports.program = program;
/**
 * Same as `program()` but with `Flags` that can be passed when the `Program` is created in order to manage initial values.
 * @category constructors
 * @since 0.5.0
 */
function programWithFlags(init, update, subscriptions) {
    if (subscriptions === void 0) { subscriptions = function () { return Sub_1.none; }; }
    return function (flags) { return program(init(flags), update, subscriptions); };
}
exports.programWithFlags = programWithFlags;
/**
 * Stops the `program` when `signal` Observable emits a value.
 * @category combinators
 * @since 0.5.4
 */
function withStop(signal) {
    return function (program) {
        var cmd$ = program.cmd$, model$ = program.model$, sub$ = program.sub$, rest = __rest(program, ["cmd$", "model$", "sub$"]);
        return __assign(__assign({}, rest), { model$: model$.pipe(operators_1.takeUntil(signal)), cmd$: cmd$.pipe(operators_1.takeUntil(signal)), sub$: sub$.pipe(operators_1.takeUntil(signal)) });
    };
}
exports.withStop = withStop;
/**
 * Runs the `Program`.
 *
 * Because the program essentially is an object of streams, "running it" means subscribing to these streams and starting to consume values.
 * @category utils
 * @since 0.5.0
 */
function run(program) {
    var dispatch = program.dispatch, cmd$ = program.cmd$, sub$ = program.sub$, model$ = program.model$;
    cmd$.subscribe(function (task) { return task().then(O.map(dispatch)); });
    sub$.subscribe(dispatch);
    return model$;
}
exports.run = run;
